import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import java.awt.*;
import java.awt.image.*;

/**
 * Die Klasse <CODE>Block</CODE> kapselt einen 8x8-Block eines Bildes. <CODE>Block</CODE> kann sich
 * vorwrts und rckwrts DCT-transformieren, dequantisieren, in einer <CODE>JTable</CODE> darstellen sowie
 * entweder per Zick-Zack-Scan oder zeilenweisem Scan in einen <CODE>BlockVektor</CODE> verwandeln. Schlielich
 * verfgt <CODE>Block</CODE> noch ber die Fhigkeit, sich als 3D-Balkendiagramm darzustellen.
 * @author  Raimund Specht, Hari
 * @see 	FloatBlock
 * @see     BlockVektor
 */
public class Block extends Matrix {

	/**
	 * Damit nicht bei jedem Aufruf von getDCT() ein neues Objekt erzeugt werden mu (was viel Zeit kostet)
	 * wird hier ein Dummy-Rckgabe Block angelegt
	 */
	FloatBlock blk=new FloatBlock();

	/**
	 * Tabelle mit vordefinierten DCT-Koeffizienten
	 */
	public final double[][] dct_coeff;// Global verfgbare, nicht berschreibbare Matrix, mit den Cosinus-Koeffizienten
	
    /**
     * Dieses Feld speichert die zugrundeliegende 8x8-Matrix als <CODE>int[][]</CODE>-Array.
     */
    public int[][] matrix;

    /**
     * Legt ein neues <CODE>Block</CODE>-Objekt an.
     * @author  Raimund Specht
     * @version 1-03.12.1999
     */
    public Block() {
        int i, j;
        matrix = new int[8][8];
        
        dct_coeff=new double[8][8];
        for (j=0; j<8; j++)
            for (i=0; i<8; i++)
                dct_coeff[i][j]=Math.cos((2*i+1)*j*Math.PI/16);
    }

    /**
     * Legt ein neues <CODE>Block</CODE>-Objekt an, wobei die Matrix mit den Werten des als Parameter bergebenen
     * <CODE>int[][]</CODE>-Arrays gefllt wird.
     * @author  Raimund Specht
     * @version 2-07.12.1999
     * @param   matrix Array, mit dessen Werten das neue <CODE>Block</CODE>-Objekt initialisiert wird.
     */
    public Block(int[][] matrix) {
        int i, j;
        
        this.matrix = new int[8][8];
        for (i=0; i<8; i++)
            for (j=0; j<8; j++)
                this.matrix[i][j] = matrix[i][j];
                
        dct_coeff=new double[8][8];
        for (j=0; j<8; j++)
            for (i=0; i<8; i++)
                dct_coeff[i][j]=Math.cos((2*i+1)*j*Math.PI/16);
    }

    /**
     * Diese Methode fhrt die (Vorwrts-)DCT-Transformation durch und liefert ein neues 
     * <CODE>FloatBlock</CODE>-Objekt mit den transformierten Werten zurck. Vollkommen neue, schnelle Version
     * @author  Christoph Ballhause
     * @version 4
     * @returns FloatBlock
     */
	public FloatBlock getDCT()
	{
		float t0, t1, t2, t3, t4, t5, t6, t7;
        
		double sq2=1/(2*Math.sqrt(2));
		int u, x, y;

		for (y=0; y<8; y++)			// process rows
		{
			t0=matrix[0][y]-128;
			t1=matrix[1][y]-128;
			t2=matrix[2][y]-128;
			t3=matrix[3][y]-128;
			t4=matrix[4][y]-128;
			t5=matrix[5][y]-128;
			t6=matrix[6][y]-128;
			t7=matrix[7][y]-128;

			blk.matrix[0][y]=(float)(sq2*(t0+t1+t2+t3+t4+t5+t6+t7));
			for (x=1; x<8; x++)
				blk.matrix[x][y]=(float)((t0*dct_coeff[0][x]+t1*dct_coeff[1][x]+t2*dct_coeff[2][x]+t3*dct_coeff[3][x]+
												  t4*dct_coeff[4][x]+t5*dct_coeff[5][x]+t6*dct_coeff[6][x]+t7*dct_coeff[7][x])/2);
		}
		
		for (x=0; x<8; x++)			// process columns
		{
			t0=blk.matrix[x][0];
			t1=blk.matrix[x][1];
			t2=blk.matrix[x][2];
			t3=blk.matrix[x][3];
			t4=blk.matrix[x][4];
			t5=blk.matrix[x][5];
			t6=blk.matrix[x][6];
			t7=blk.matrix[x][7];

			blk.matrix[x][0]=(float)(sq2*(t0+t1+t2+t3+t4+t5+t6+t7));
			for (y=1; y<8; y++)
				blk.matrix[x][y]=(float)((t0*dct_coeff[0][y]+t1*dct_coeff[1][y]+t2*dct_coeff[2][y]+t3*dct_coeff[3][y]+
												  t4*dct_coeff[4][y]+t5*dct_coeff[5][y]+t6*dct_coeff[6][y]+t7*dct_coeff[7][y])/2);
		}
		return blk;
	}
	
    /**
     * Diese Methode fhrt die inverse DCT-Transformation durch und liefert ein neues 
     * <CODE>Block</CODE>-Objekt mit den rcktransformierten Werten zurck. Schnelle Version der iDCT
     * @author  Raimund Specht, Hari, Christoph Ballhause
     * @version 3
     * @returns Block
     */
	public Block getIDCT()
	{
		float t0, t1, t2, t3, t4, t5, t6, t7;
        
		double sq2=1/Math.sqrt(2);
		int u, x, y;
		
		for (y=0; y<8; y++)			// process rows
		{
			t0=matrix[0][y];
			t1=matrix[1][y];
			t2=matrix[2][y];
			t3=matrix[3][y];
			t4=matrix[4][y];
			t5=matrix[5][y];
			t6=matrix[6][y];
			t7=matrix[7][y];
			
		     for (x=0; x<8; x++)
				blk.matrix[x][y]=(float)((t0*dct_coeff[x][0]*sq2+t1*dct_coeff[x][1]+t2*dct_coeff[x][2]+t3*dct_coeff[x][3]+
			  						  t4*dct_coeff[x][4]+t5*dct_coeff[x][5]+t6*dct_coeff[x][6]+t7*dct_coeff[x][7])/2);
		}
		
		for (x=0; x<8; x++)			// process columns
		{
			t0=blk.matrix[x][0];
			t1=blk.matrix[x][1];
			t2=blk.matrix[x][2];
			t3=blk.matrix[x][3];
			t4=blk.matrix[x][4];
			t5=blk.matrix[x][5];
			t6=blk.matrix[x][6];
			t7=blk.matrix[x][7];
			
			for (y=0; y<8; y++)
				blk.matrix[x][y]=(float)((t0*dct_coeff[y][0]*sq2+t1*dct_coeff[y][1]+t2*dct_coeff[y][2]+t3*dct_coeff[y][3]+
												  t4*dct_coeff[y][4]+t5*dct_coeff[y][5]+t6*dct_coeff[y][6]+t7*dct_coeff[y][7])/2)+128;
		}

		return blk.toIntBlock();
	}


    /**
     * Diese Methode liefert den Block als "linearisierten" 1x64-<CODE>BlockVektor</CODE> zurck. Dabei
     * werden die Werte des 8x8-Blocks in der bekannten JPEG-"Zick-Zack"-Reihenfolge ausgelesen.
     * @author  Raimund Specht
     * @version 1-13.12.1999
     * @returns BlockVektor
     * @see     BlockVektor
     */
    public BlockVektor getZigZagScan() {
        final int[] zickZackX = {0, 1, 0, 0, 1, 2, 3, 2, 1, 0, 0,
                                 1, 2, 3, 4, 5, 4, 3, 2, 1, 0, 0,
                                 1, 2, 3, 4, 5, 6, 7, 6, 5, 4, 3, 2, 1, 0,
                                 1, 2, 3, 4, 5, 6, 7, 7, 6, 5, 4, 3, 2,
                                 3, 4, 5, 6, 7, 7, 6, 5, 4, 5, 6, 7,
                                 7, 6, 7};
        final int[] zickZackY = {0, 0, 1, 2, 1, 0, 0, 1, 2, 3, 4,
                                 3, 2, 1, 0, 0, 1, 2, 3, 4, 5, 6,
                                 5, 4, 3, 2, 1, 0, 0, 1, 2, 3, 4, 5, 6, 7, 7,
                                 6, 5, 4, 3, 2, 1, 2, 3, 4, 5, 6, 7, 7,
                                 6, 5, 4, 3, 4, 5, 6, 7, 7, 6, 5, 6, 7, 7};
        BlockVektor bv;
        int i;
        
        bv = new BlockVektor();
        for (i=0; i<64; i++)
            bv.werte[i] = (int)matrix[zickZackX[i]][zickZackY[i]];
        return bv;
    }

    /**
     * Diese Methode liefert den Block als "linearisierten" 1x64-<CODE>BlockVektor</CODE> zurck. Dabei
     * werden die Werte des 8x8-Blocks von links nach rechts und von oben nach unten ausgelesen.
     * @author  Raimund Specht
     * @version 1-13.12.1999
     * @returns BlockVektor
     * @see     BlockVektor
     */
    public BlockVektor getLinearScan() {
        BlockVektor bv;
        int x, y;
        
        bv = new BlockVektor();
        for (y=0; y<8; y++)
            for (x=0; x<8; x++)
                bv.werte[x + 8*y] = (int)matrix[x][y];
        
        return bv;
    }

    /**
     * Diese Methode macht die Quantisierung des Blocks rckgngig (Dequantisierung). Es wird ein neues 
     * <CODE>Block</CODE>-Objekt zurckgeliefert, das ursprngliche Objekt bleibt unverndert.
     * @author  Raimund Specht
     * @version 1-03.12.1999
     * @param   qMatrix Die zu verwendende "Dequantisierungsmatrix"
     * @returns Block
     */
    public Block getDequantisiert(Matrix qMatrix) {
        int i, j;
        Block blk = new Block();

        for (i=0; i<8; i++)
            for (j=0; j<8; j++)
                blk.matrix[i][j] = matrix[i][j] * qMatrix.matrix[i][j];
        return blk;
    }

    /**
     * Zeichnet die Werte des Blocks als 3D-Balkendiagramm. Dabei werden die Absolut-Betrge
     * der jeweiligen Werte verwendet. Der betragsmig grte Wert im Block wird als Balken
     * mit maximaler Ausdehnung dargestellt, alle anderen Balken werden in der Gre entsprechend
     * angepat.
     * @author  Christoph Ballhause
     * @version 1-09.12.2k-1
     * @param   width Gewnschte Breite des erzeugten 3D-Balkendiagramms
     * @param   height Gewnschte Hhe des erzeugten 3D-Balkendiagramms
     * @param   Color Gewnschte Farbe der Balken
     * @returns java.awt.image.BufferedImage
     */
    public BufferedImage paintBar3D(int width, int height, Color c)
    {
        BufferedImage dummy=new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
        Graphics2D g=dummy.createGraphics();
        int x, y;
        float bh, bx, by;
        float bsx, bsy, bbx, bby, bsl;
        float ratio;
        Polygon p;
        
        // Berechnung der Parameter zur perspektivischen Darstellung
        bbx=(width-height*GUIParameter.DIAG_RATIO_XY*GUIParameter.DIAG_RATIO_Y)/(8/GUIParameter.DIAG_WIDTH);// Breite eines Balkens
        bby=(height*GUIParameter.DIAG_RATIO_Y)/(8/GUIParameter.DIAG_HEIGHT);            // Perspektivische Tiefe eines Balkens
        bsx=(width-height*GUIParameter.DIAG_RATIO_XY*GUIParameter.DIAG_RATIO_Y)/8;    // Breite eines Balkens inkl. Zwischenraum
        bsy=(height*GUIParameter.DIAG_RATIO_Y)/8;                                                // Tiefe eines Balkens inkl. Zwischenraum
        bsl=(height*GUIParameter.DIAG_RATIO_XY*GUIParameter.DIAG_RATIO_Y)/8;            // Parallelverschiebung pro Y-Einheit
        
        g.setColor(GUIParameter.INACTIVE_BACKGROUND);
        g.fillRect(0, 0, width, height);
        
        ratio=Math.abs(matrix[0][0]);        // betragsmig grten Wert bestimmen
        for (y=0; y<8; y++)
            for (x=0; x<8; x++)
                if (Math.abs(matrix[x][y])>ratio)
                    ratio=Math.abs(matrix[x][y]);
        
        if (ratio>0)								// wenn die Matrix nur aus Nullen besteht, sind alle Balken 0
        		ratio=(height-8*bsy)/ratio;
        
        for (y=0; y<8; y++)                // Reihen und Spalten von hinten nach vorn und links nach rechts durchlaufen
            for (x=0; x<8; x++)
            {
                bh=Math.abs(matrix[x][y])*ratio;                // Hhe in Pixeln berechnen
                bx=bsl*(7-y)+bsx*x;
                by=height-bsy*(7-y);
                
                p=new Polygon();
                p.addPoint(Math.round(bx), Math.round(by));                    // Frontseite malen
                p.addPoint(Math.round(bx+bbx), Math.round(by));
                p.addPoint(Math.round(bx+bbx), Math.round(by-bh));
                p.addPoint(Math.round(bx), Math.round(by-bh));
                g.setColor(c);
                g.fillPolygon(p);
                g.setColor(Color.black);
                g.drawPolygon(p);
                p=new Polygon();                        // Oberseite malen
                p.addPoint(Math.round(bx), Math.round(by-bh));
                p.addPoint(Math.round(bx+bbx), Math.round(by-bh));
                p.addPoint(Math.round(bx+bbx+bsl), Math.round(by-bh-bby));
                p.addPoint(Math.round(bx+bsl), Math.round(by-bh-bby));
                g.setColor(c.brighter());
                g.fillPolygon(p);
                g.setColor(Color.black);
                g.drawPolygon(p);
                p=new Polygon();                        // Seitenwand malen
                p.addPoint(Math.round(bx+bbx), Math.round(by));
                p.addPoint(Math.round(bx+bbx+bsl), Math.round(by-bby));
                p.addPoint(Math.round(bx+bbx+bsl), Math.round(by-bh-bby));
                p.addPoint(Math.round(bx+bbx), Math.round(by-bh));
                g.setColor(c.darker());
                g.fillPolygon(p);
                g.setColor(Color.black);
                g.drawPolygon(p);
            }
        
        return dummy;
    }

    /**
     * Implementiert das von <CODE>Matrix</CODE> geerbte Interface <CODE>TableModel</CODE>.
     * @author  Raimund Specht
     * @version 1-03.12.1999
     * @param   rowIndex Zeile der angesprochenen Zelle
     * @param   columnIndex Spalte der angesprochenen Zelle
     * @returns Object
     * @see     Matrix
     * @see     javax.swing.table.TableModel
     */
    public Object getValueAt(int rowIndex, int columnIndex) { 
        return new Integer((int)matrix[columnIndex][rowIndex]);
    }

    /**
     * Mit diese Methode kann fr eine beliebige Zelle berprft werden, ob diese Zelle vom Benutzer
     * verndert werden kann, wenn der Block in einer <CODE>JTable</CODE> dargestellt wird. Die Klasse
     * <CODE>Block</CODE> liefert hier immer <CODE><B>false</B></CODE> zurck, d.h. es kann keine Zelle
     * bearbeitet werden.
     * @author  Raimund Specht
     * @version 1-03.12.1999
     * @param   rowIndex Zeile der angesprochenen Zelle
     * @param   columnIndex Spalte der angesprochenen Zelle
     * @returns boolean
     * @see     javax.swing.table.TableModel
     */
    public boolean isCellEditable(int rowIndex, int columnIndex) {
        return false;
    }

    /**
     *
     * @author  Raimund Specht
     * @version 1-02.12.1999
     * @param   color Die gewnschte Hintergrundfarbe der <CODE>JTable</CODE>
     * @returns java.swing.JTable
     */
    public JTable paintMatrix(Color farbe) {
        JTable table;
        
        table = new JTable(this);
        table.setBackground(farbe);
        table.setFont(GUIParameter.SMALL_FONT);
        table.setPreferredSize(new Dimension(192, table.getPreferredSize().height));
        table.setMaximumSize(new Dimension(192, table.getPreferredSize().height));
        table.setMinimumSize(new Dimension(192, table.getPreferredSize().height));
        return table;
    }
}