import java.util.*;
import java.io.*;

/**
 * Die Klasse <CODE>HuffmanWriter</CODE> stellt einen Puffer dar, in den man
 * Huffman Symbole rein schreiben kann, und zwar mit den Methoden
 * <CODE>writeSymbol</CODE>, <CODE>writeByte</CODE> und <CODE>writeBytes</CODE>.
 * Die Methode <CODE>getBitsWritten</CODE> sagt, wie viele Bits man schon in
 * den Puffer gestopft hat. Spter kann man dann den ganzen Puffer mit
 * <CODE>writeToFile</CODE> in eine Datei schreiben. Vorher muss man aber
 * die Methode <CODE>flush</CODE> aufrufen, die das letzte (unvollstndige)
 * Byte mit Bits auffllt und an den Puffer anhngt.
 * Ich htte die Klasse vielleicht besser "HuffmanBuffer" nennen sollen,
 * aber jetzt nder ich das auch nicht mehr.
 * Der Puffer besteht brigens aus einem <CODE>Vector</CODE> von vielen
 * kleinen Puffern der Gre <CODE>maxBufferSize</CODE>.
 * @author  Hari
 * @version 1 - 4.2.00
 */
public class HuffmanWriter {
    private int currentVal;
    private int currentSize;
    private Vector outBuffers;
    private final int maxBufferSize = 4096;
    private byte currentBuffer[];
    private int currentBufferSize;
    
    /**
     * Erzeugt einen HuffmanWriter mit leerem Puffer
     * @author  Hari
     * @version 1 - 4.2.00
     */
    public HuffmanWriter() {
        currentVal = 0;
        currentSize = 0;    
        outBuffers = new Vector();
        currentBuffer = new byte[maxBufferSize];
        outBuffers.addElement(currentBuffer);
        currentBufferSize = 0;
    }
    
    /**
     * Liefert die Anzahl der bisher in den Puffer geschriebenen Bits zurck
     * @author  Hari
     * @version 1 - 4.2.00
     * returns int
     */
    public int getBitsWritten() {
        return (outBuffers.size()-1)*maxBufferSize*8 + currentBufferSize*8 + currentSize;
    }
    
    /**
     * Schreibt ein <CODE>byte</CODE> ans Ende des Puffers.
     * Vorsicht! Es wird kein <CODE>flush</CODE> durchgefhrt!
     * @author  Hari
     * @version 1 - 4.2.00
     * @param byte
     */
    public void writeByte(byte b) {
        if (currentBufferSize==maxBufferSize) {
            currentBuffer = new byte[maxBufferSize];
            currentBufferSize = 0;
            outBuffers.addElement(currentBuffer);
        }
        currentBuffer[currentBufferSize] = b;
        currentBufferSize++;
    }
    
    /**
     * Schreibt ein <CODE>byte[]</CODE> ans Ende des Puffers.
     * Vorsicht! Es wird kein <CODE>flush</CODE> durchgefhrt!
     * @author  Hari
     * @version 1 - 4.2.00
     * @param byte[]
     */
    public void writeBytes(byte b[]) {
        for (int i=0; i<b.length; i++)
            writeByte(b[i]);
    }
    
    /**
     * Schreibt die ersten <CODE>length</CODE> Elemente eines
     * <CODE>byte[]</CODE> ans Ende des Puffers.
     * Vorsicht! Es wird kein <CODE>flush</CODE> durchgefhrt!
     * @author  Hari
     * @version 1 - 4.2.00
     * @param byte[]
     * @param int
     */
    public void writeBytes(byte b[], int length) {
        for (int i=0; i<length; i++)
            writeByte(b[i]);
    }
    
    /**
     * Schreibt ein <CODE>byte</CODE> ans Ende des Puffers und hngt
     * noch ein <CODE>0</CODE> Byte an, falls es sich um ein <CODE>0xFF</CODE>
     * handelt (so wills das JPEG Format).
     * Wird von writeSymbol aufgerufen, deshalb mssen wir den Puffer nicht
     * flushen.
     * @author  Hari
     * @version 1 - 4.2.00
     * @param byte
     */
    private void outputCodeByte(byte b) {
        if (b==(byte)0xFF) {  // aha! ein 0xff, da mssen wir noch ein 0 byte hinterschreiben
            writeByte(b);
            writeByte((byte)0);
        }
        else
            writeByte(b);
    }
            
    /**
     * Schreibt ein Huffman Symbol ans Ende des Puffers.
     * Das Symbol wird durch die Lnge (<CODE>size</CODE>) und
     * die Bitfolge in <CODE>val</CODE> angegeben.
     * @author  Hari
     * @version 1 - 4.2.00
     * @param int
     * @param int
     */
    public void writeSymbol(int size, int val) {
        currentVal = val + (currentVal<<size);
        currentSize += size;
        while (currentSize>=8) {
            currentSize -= 8;
            int output = currentVal>>currentSize;
            outputCodeByte((byte)output);
            currentVal -= output<<currentSize;
        }
    }
    
    /**
     * Schreibt das letzte, noch nicht vollstndig mit Bits aufgefllte
     * Byte ans Ende des Puffers.
     * Vor <CODE>writeToFile</CODE> ist es gut, das zu machen.
     * @author  Hari
     * @version 1 - 4.2.00
     */
    public void flush() {
        writeByte((byte)(currentVal<<(8-currentSize)));
        currentVal=0;
        currentSize = 0;
    }
    
    /**
     * Schreibt den ganzen Puffer in die Datei mit Namen <CODE>filename</CODE>.
     * Wenn es die Datei schon gibt, wird sie rcksichtslos berschrieben.
     * Vorher <CODE>flush</CODE> aufrufen! (wie schon mehrmals erwhnt)
     * @author  Hari
     * @version 1 - 4.2.00
     * @param String
     */
    public boolean writeToFile(String filename) {
        File outFile = new File(filename);
        FileOutputStream fStream = null;
        try {
            fStream = new FileOutputStream(outFile);
        } catch (IOException e) {
            e.printStackTrace();
            return false;
        }
        BufferedOutputStream outStream = new BufferedOutputStream(fStream);
        for (int i=0; i<outBuffers.size()-1; i++) {
            try {
                outStream.write((byte[])outBuffers.elementAt(i), 0, maxBufferSize);
            } catch (IOException e) {
                e.printStackTrace();
                return false;
            }
        }
        try {
            outStream.write((byte[])outBuffers.elementAt(outBuffers.size()-1), 0, currentBufferSize);
        } catch (IOException e) {
            e.printStackTrace();
            return false;
        }
        
        try {
            outStream.flush();
        } catch (IOException e) {
            e.printStackTrace();
            return false;
        }
        try {
            outStream.close();
        } catch (IOException e) {
            e.printStackTrace();
            return false;
        }
        return true;
    }
}