import java.awt.*;
import java.awt.image.*;
import java.util.*;

/**
 * Diese Klasse stellt einen Bildkanal in Symbolform dar.
 * Das heit, es werden alle Symbolblcke des Kanals in einem
 * <CODE>Vector</CODE> gespeichert.
 * @author Hari
 * @see SymbolBlock
 */
public class SymbBlockFolge {
    private Vector bloecke;    
    
    public int lengthVLICoded;

   /**
    * Konstruktor, erzeugt leere Folge
    * @author Hari
    * @version 1
    */
    public SymbBlockFolge() {
        bloecke = new Vector();
    }
    
   /**
    * Gibt den <CODE>i</CODE>-ten <CODE>SymbolBlock</CODE> zurck
    * @author Hari
    * @version 1
    */
    public SymbolBlock elementAt(int i) {
        return (SymbolBlock)bloecke.elementAt(i);
    }
    
   /**
    * Fgt einen <CODE>SymbolBlock</CODE> hinzu
    * @author Hari
    * @version 1
    */
    public void addElement(SymbolBlock block) {
        bloecke.addElement(block);
    }
    
   /**
    * Liefert die Gesamtlnge der Symbole aller Blcke nach der RLE Kodierung
    * zurck (nach dem Erzeugen eines <CODE>SymbolBlocks</CODE> ist dieser
    * bereits von Natur aus RLE kodiert)
    * @author Hari
    * @version 1
    */
    public int getLengthRLECoded() {
        int length = 0;
        for (int i=0; i<bloecke.size(); i++)
            length += elementAt(i).getLengthRLECoded();
        return length;
    }
    
    /**
     * Interpretiert die RLESymbolfolge als Folge von DC Werten und DPCM-kodiert sie.
     * Es wird vorausgesetzt, dass immer abwechselnd ein <CODE>RLESymbol1</CODE>
     * und ein <CODE>RLESymbol2</CODE> kommt.
     * Beim <CODE>RLESymbol1</CODE> wird hier der <CODE>zeroCount</CODE> ignoriert,
     * bzw. nicht verwendet.
     * Die Methode setzt die <CODE>lengthDPCMCoded</CODE>.
     * @author Hari
     * @version 2 - 2.1.00
     */
    public void doDPCMCoding() {
        int ampl, prevAmpl, i, code;
        SymbolBlock block;
        
        prevAmpl = 0; // bruchten wir nicht, aber der piss Compiler denkt er wre schlauer als ich und besteht auf einem init Wert
        for (i=0; i<bloecke.size(); i++) {
            block = (SymbolBlock)bloecke.elementAt(i);
            ampl = block.getDCVal();
            if (i>0) {
                code = ampl - prevAmpl;    // Werte nach dem Ersten: Differenz codieren
                block.setDCVal(code);
            }
            else
                code = ampl;              // Erster Wert: Amplitude codieren
            prevAmpl= ampl;
        }
    }
    
   /**
    * Fhrt fr jeden Block die VLI Kodierung durch, indem
    * die entsprechende gleichnamige Methode des Blocks aufgerufen wird.
    * @author Hari
    * @version 1
    */
    public void doVLICoding() {
        int i;
        
        lengthVLICoded = 0;
        for (i=0; i<bloecke.size(); i++) {
            SymbolBlock block = elementAt(i);
            block.doVLICoding();
            lengthVLICoded += block.lengthVLICoded;
        }
    }
    
   /**
    * Fhrt fr den Block mit Index <CODE>blockIdx</CODE> die Huffman Kodierung
    * durch, indem die entsprechende gleichnamige Methode des Blocks aufgerufen
    * wird.
    * @author Hari
    * @version 1
    * @param int blockIdx
    * @param OurHuffmanTable dcHuffmanTable
    * @param OurHuffmanTable acHuffmanTable
    * @param HuffmanWriter hWriter
    */
    public void huffmanCodeBlock(int blockIdx, OurHuffmanTable dcHuffmanTable, OurHuffmanTable acHuffmanTable, HuffmanWriter hWriter) {
        SymbolBlock block = elementAt(blockIdx);
        block.doHuffmanCoding(dcHuffmanTable, acHuffmanTable, hWriter);
    }
    
   /**
    * Liefert ein Histogramm ber die Lngen der Symbole aller Blocke
    * nach der VLI Kodierung zurck
    * @author Hari
    * @version 1
    * @returns Histogramm
    */
    public Histogramm getVLIHistogramm() {
        int i;
        Histogramm histo = new Histogramm(16);  // 11 bits fr VLI und 4 fr Lnge, also 0..15
        for (i=0; i<bloecke.size(); i++)
            histo.addHisto(elementAt(i).getVLIHistogramm());
        return histo;
    }
    
   /**
    * Liefert ein Histogramm ber die Lngen der VLI Werte der DC Symbole
    * zurck.
    * @author Hari
    * @version 1
    * @returns Histogramm
    */
    public Histogramm getDCFreqHisto() {
        Histogramm histo;
        int i;
        
        histo = new Histogramm(257);
        for (i=0; i<bloecke.size(); i++) {
            int dcSize = elementAt(i).getDCSize();
            histo.werte[dcSize]++;
        }
        histo.werte[256] = 1;  // fr Bauung der Huffman Tabelle mssen wir das so setzen
        return histo;
    }
    
   /**
    * Liefert ein Histogramm ber die <CODE>(run,size)</CODE> Paare zurck,
    * indem die gleichnamige Methode von <CODE>SymbolBlock</CODE> aufgerufen
    * wird und die Histogramme addiert werden.
    * @author Hari
    * @version 1
    * @returns Histogramm
    */
    public Histogramm getACFreqHisto() {
        Histogramm histo;
        int i;
        
        histo = new Histogramm(257);
        for (i=0; i<bloecke.size(); i++)
            histo.addHisto(elementAt(i).getACFreqHisto());
        histo.werte[256] = 1;  // fr Bauung der Huffman Tabelle mssen wir das so setzen
        return histo;
    }
    
   /**
    * Liefert ein Histogramm ber die Lngen der Huffman Symbole aller Blcke,
    * indem die gleichnamige Methode von <CODE>SymbolBlock</CODE> aufgerufen
    * wird und die Histogramme addiert werden.
    * @author Hari
    * @version 1
    * @returns Histogramm
    */
    public Histogramm getHuffmanHistogramm() {
        Histogramm histo;
        int i;
        
        histo = new Histogramm(17);  // Huffman Code ist auf 16 bits beschrnkt, also brauchen wir den bereich 0-16
        for (i=0; i<bloecke.size(); i++)
            histo.addHisto(elementAt(i).getHuffmanHistogramm());
        return histo;
    }
}