import java.awt.*;
import java.awt.image.*;
import java.util.*;

/**
 * Ein <CODE>SymbolBlock</CODE> ist ein 8x8 Block in "Symbolform",
 * das heit, er wird nicht durch 64 Zahlenwerte, sondern durch
 * zwei <CODE>RLESymbol</CODE> fr den DC Wert (Das erste fr size und
 * das zweite fr den DPCM Wert) und eine <CODE>RLESymbolfolge</CODE>
 * fr die AC Werte beschrieben.
 * Auf dieser Klasse ist die VLI- und Huffman Kodierung implementiert
 * (<CODE>doVLICoding</CODE> und <CODE>doHuffmanCoding</CODE>)
 * Fernerhin kann man sich mit <CODE>paintHuffmanSizes</CODE> ein
 * Balkendiagramm der Huffman Lngen geben lassen.
 * <CODE>getACFreqHisto</CODE> gibt eine Hufigkeitstabelle fr die
 * Erstellung der Huffman Tabelle zurck, und mit <CODE>getDCVal</CODE>
 * und <CODE>getDCSize</CODE> kann man sich das DC Symbol geben lassen.
 * <CODE>getHuffmanHistogramm</CODE> gibt eine Hufigkeitstabelle ber die
 * Bitlngen der Huffman Symbole zurck, mit denen der Block kodiert wurd.
 * Der Unterschied zu <CODE>getACFreqHisto</CODE> ist, dass bei letzerem
 * die uncodierten Symbole betrachtet werden, also die Werte
 * <CODE>run*16+size)</CODE>.
 * @author Hari
 */
public class SymbolBlock {
    private RLESymbol1 dcSymb1;
    private RLESymbol2 dcSymb2;
    private RLESymbolfolge acSymbole;
    
    public int lengthVLICoded;
    
   /**
    * Konstruktor, erzeugt leeren Block
    * @author Hari
    * @version 1
    * @returns SymbolBlock
    */
    public SymbolBlock() {
        dcSymb1 = new RLESymbol1();
        dcSymb2 = new RLESymbol2();
        acSymbole = new RLESymbolfolge();
    }

   /**
    * Konstruktor, liefert Block mit dem DC Wert <CODE>amplitude</CODE>
    * und den AC Symbolen aus <CODE>acSymbFolge</CODE>
    * @author Hari
    * @version 1
    * @param int
    * @param RLESymbolfolge
    * @returns SymbolBlock
    */
    public SymbolBlock(int amplitude, RLESymbolfolge acSymbFolge) {
        dcSymb1 = new RLESymbol1();
        dcSymb2 = new RLESymbol2();
        dcSymb2.amplitude = amplitude;
        acSymbole = acSymbFolge;
    }

   /**
    * Liefert den DC Wert zurck, bzw. nach der VLI Kodierung dessen VLI Symbol
    * @author Hari
    * @version 1
    * @returns int
    */
    public int getDCVal() {
        return dcSymb2.amplitude;
    }
    
   /**
    * Liefert die Lnge des VLI Symbols fr den DC Wert
    * @author Hari
    * @version 1
    * @returns int
    */
    public int getDCSize() {
        return dcSymb1.sizeNonZero;
    }    
        
   /**
    * Setzt den DC Wert auf <CODE>amplitude</CODE>
    * Wird bei der DPCM Kodierung aufgerufen, um den DC Wert durch die
    * Differenz zum Vorhergehenden zu ersetzen.
    * @author Hari
    * @version 1
    * @param int
    */
    public void setDCVal(int amplitude) {
        dcSymb2.amplitude = amplitude;
    }
    
   /**
    * Liefert die Lngen aller Symbole des Blocks (DC & AC) als <CODE>int[]</CODE>
    * zurck. Wird von <CODE>paintHuffmanSizes</CODE> benutzt.
    * @author Hari
    * @version 1
    * @returns int[]
    */
    private int[] getSizes() {
        Vector vec;
        int intValues[];
        int i;
        
        vec = new Vector();
        vec.addElement(new Integer(dcSymb1.huffSize));
        for (i=0; i<acSymbole.size(); i++) {
            RLESymbol symb = acSymbole.elementAt(i);
            if (symb instanceof RLESymbol1)
                vec.addElement(new Integer(((RLESymbol1)symb).huffSize));
        }
        intValues = new int[vec.size()];
        for (i=0; i<vec.size(); i++)
            intValues[i] = ((Integer)vec.elementAt(i)).intValue();
        return intValues;
    }
    
    /**
     * Liefert ein Balkendiagramm mit den Lngen der Huffman Symbole zurck.
     * Die Durchschnittswerte vor und nach der Huffman Kodierung werden als
     * zwei verschieden farbige Linien eingezeichnet.
     * @author Hari
     * @version 1
     * @param int width
     * @param int height
     * @param Color colorDC
     * @param Color colorAC
     * @param Color colorAvgBefore
     * @param Color colorAvgAfter
     * @returns BufferedImage
     */
    public BufferedImage paintHuffmanSizes(int width, int height, Color colorDC, Color colorAC, Color colorAvgBefore, Color colorAvgAfter) {
        BufferedImage image;
        Graphics2D g2D;
        FontMetrics fm;
        int werte[];
        int i, x, y;
        int balkenBreite, luecke;
        float faktor;
        int randUnten = 20;
        int randOben = 5;
        int randLinx;
        int summeLaengen;
        
        werte = getSizes();
        
        image = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
        g2D = image.createGraphics();
        g2D.setColor(GUIParameter.INACTIVE_BACKGROUND);
        g2D.fillRect(0, 0, width, height);
        fm = g2D.getFontMetrics();
        
        // beschriftung linx
        int numYStringz = 5;
        String yStringz[] = new String[numYStringz];
        yStringz[0] = new String("0");
        yStringz[1] = new String("4");
        yStringz[2] = new String("8");
        yStringz[3] = new String("12");
        yStringz[4] = new String("16");
        
        // Maximum ber Stringlngen bilden und randLinx zuweisen
        randLinx = 0;
        for (i=0; i<numYStringz; i++) {
            int strWidth = fm.stringWidth(yStringz[i]);
            if (strWidth>randLinx)
                randLinx = strWidth;
        }
        
        // Linien und Beschriftung fr y Axe machen
        randLinx += 5+2+5+2;  // Abstand Linker Rand des g2D - y Zahl
                              // + Abstand Zahl - Markierungsstrich fr Zahl
                              // + Breite Markierungsstrich
                              // + Abstand y Axe - erster Balken
        g2D.setColor(Color.black);
        g2D.drawLine(randLinx-2-1, randOben, randLinx-2-1, height-randUnten);
        for (i=0; i<numYStringz; i++) {
            y = height-randUnten - (height-randUnten-randOben)*i/(numYStringz-1);
            g2D.drawLine(randLinx-5, y, randLinx-2-1, y);
            g2D.drawString(yStringz[i], randLinx-2-fm.stringWidth(yStringz[i])-2-5, y+fm.getHeight()/2);
        }
        
        faktor = (float)(height-randOben-randUnten) / 16;
        balkenBreite = (width-randLinx) / (2*werte.length);
        if (balkenBreite>10)
            balkenBreite=10;
        luecke = (width-randLinx) / (2*werte.length);
        if (luecke>10)
            luecke=10;
        
        g2D.drawLine(randLinx, height-randUnten, width-1, height-randUnten);
        
        // balken zeichnen
        x = randLinx;
        for (i=0; i<werte.length; i++) {
            Polygon poly = new Polygon();
            String s;
            
            poly.addPoint(x, height-randUnten);
            poly.addPoint(x+balkenBreite-1, height-randUnten);
            poly.addPoint(x+balkenBreite-1, height-randUnten-(int)(werte[i]*faktor));
            poly.addPoint(x, height-randUnten-(int)(werte[i]*faktor));
            if (i==0)
                g2D.setColor(colorDC);
            else
                g2D.setColor(colorAC);
            g2D.fillPolygon(poly);
            g2D.setColor(Color.black);
            g2D.drawPolygon(poly);
            x += balkenBreite+luecke;
        }
        
        // Durchschnittslnge vorher und nachher als Linie zeichnen
        summeLaengen = 4 + 8*(werte.length-1);
        y = height-randUnten-(int)(faktor*summeLaengen/werte.length);
        g2D.setColor(colorAvgBefore);
        g2D.drawLine(randLinx, y, width-1, y);

        summeLaengen = 0;
        for (i=0; i<werte.length; i++)
            summeLaengen += werte[i];
        y = height-randUnten-(int)(faktor*summeLaengen/werte.length);
        g2D.setColor(colorAvgAfter);
        g2D.drawLine(randLinx, y, width-1, y);
        
        // Erklrungstext zeichnen
        String str;
        int strHeight = fm.getHeight();
        
        g2D.setColor(colorDC);
        x = randLinx+5;
        g2D.fillRect(x, height-5-strHeight/2-2, balkenBreite, strHeight/2);
        x += balkenBreite+5;
        g2D.setColor(Color.black);
        str = new String("DC");
        g2D.drawString(str, x, height-5);
        x += fm.stringWidth(str) + 10;
        
        g2D.setColor(colorAC);
        g2D.fillRect(x, height-5-strHeight/2-2, balkenBreite, strHeight/2);
        x += balkenBreite+5;
        g2D.setColor(Color.black);
        str = new String("AC");
        g2D.drawString(str, x, height-5);
        x += fm.stringWidth(str) + 10;

        g2D.setColor(colorAvgBefore);
        g2D.drawLine(x, height-5-strHeight/2, x+30, height-5-strHeight/2);
        g2D.setColor(Color.black);
        x += 30+5;
        str = new String("Durchschn. Lnge vor Huffman");
        g2D.drawString(str, x, height-5);
        x += fm.stringWidth(str) + 10;
        
        g2D.setColor(colorAvgAfter);
        g2D.drawLine(x, height-5-strHeight/2, x+30, height-5-strHeight/2);
        g2D.setColor(Color.black);
        x += 30+5;
        str = new String("nach Huffman");
        g2D.drawString(str, x, height-5);
        
        return image;
    }
    
   /**
    * Liefert ein Histogramm ber die <CODE>(run,size)</CODE> Paare zurck,
    * genauer gesagt, ber die Werte <CODE>run*16+size</CODE>.
    * @author Hari
    * @version 1
    * @returns Histogramm
    */
    public Histogramm getACFreqHisto() {
        int i;
        Histogramm histo;
        
        histo = new Histogramm(257);  // zum HuffTable bauen brauchen wir noch einen zustzlichen 256. Wert
        for (i=0; i<acSymbole.size(); i++)
            if (acSymbole.elementAt(i) instanceof RLESymbol1) {
                RLESymbol1 symb = (RLESymbol1)acSymbole.elementAt(i);
                histo.werte[symb.zeroCount*16 + symb.sizeNonZero]++;
            }
        return histo;
    }
        
   /**
    * Spuckt die (eigentlich hypothetische) Lnge des Blocks nach der
    * RLE Kodierung aus.
    * Dient zur Veranschaulichung des Kompressionsfaktors im DPCM/VLI Schritt
    * @author Hari
    * @version 1
    * @returns int
    */
    public int getLengthRLECoded() {
        int i, size;
        
        // dc
        size = 8+11;   // dcSymb1 + dcSymb2
        
        // ac
        for (i=0; i<acSymbole.size(); i++)
            if (acSymbole.elementAt(i) instanceof RLESymbol2)
                size += 10;  // amplituden der ac komponenten haben 10 bit
            else
                size += 8;  // 4 fr size, 4 fr runlength
        return size;
    }
        
   /**
    * Liefert Gre eines unkodierten Wertes nach der VLI Kodierung,
    * also so quasi die Anzahl signifikanter Bits des Betrages eines Integers
    * @author Hari
    * @version 1
    * @param int
    * @returns int
    */
    private int getVLISize(int val) {
        int bitLength;

        bitLength = 0;
        if (val<0)
            val = -val;
        while (1<<bitLength <= val)
            bitLength++;
        return bitLength;
    }
     
    /**
     * Liefert den VLI Wert eines integers (ohne die Lngeninformation)
     * @author Hari
     * @version 1
     * @param int
     * @returns int
     */
     private int getVLIValue(int i) {
        int bits;
        int res;
        
        if (i==0)
            return 0;
              
        if (i>0)
            res = i;
        else {
            int size = 0;
            while (1<<size <= -i)
                size++;
            res = i + (1<<size) - 1;
        }
        return res;
     }
     
    /**
     * VLI-kodiert die <CODE>RLESymbol2</CODE>'s in der Folge und setzt die
     * <CODE>sizeNonZero</CODE> im vorhergehenden <CODE>RLESymbol1</CODE>.
     * Wir knnen davon ausgehen, dass vor jedem <CODE>RLESymbol2</CODE>
     * ein <CODE>RLESsymbol1</CODE> kommt.
     * Die VLI codierten Werte sind innerhalb des integers rechzbndig angeordnet,
     * also vom niederwertigsten Bit her.
     * Die Methode setzt die <CODE>lengthVLICoded</CODE>.
     * @author Hari
     * @version 2 - 28.12.99
     */
    public void doVLICoding() {
        RLESymbol symb;
        RLESymbol1 symb1;
        RLESymbol2 symb2;
        int i;
        int size, val;
        
        // dc
        size = getVLISize(dcSymb2.amplitude);
        lengthVLICoded = size+4;   // plus die 4 Bits fr die Lngeninfo
        dcSymb1.sizeNonZero = size;
        dcSymb2.amplitude = getVLIValue(dcSymb2.amplitude);  // symbol2 wird VLI codiert
            
        // ac
        for (i=0; i<acSymbole.size(); i++) {
            symb = acSymbole.elementAt(i);
            if (symb instanceof RLESymbol2) {
                symb2 = (RLESymbol2)symb;
                size = getVLISize(symb2.amplitude);
                
                symb1 = (RLESymbol1)acSymbole.elementAt(i-1);   // Vorgnger-RLESymbol ist ein symb1
                symb1.sizeNonZero = size;  // im Vorgnger-RLESymbol sizeNonZero setzen
                    
                val = getVLIValue(symb2.amplitude);
                symb2.amplitude = val;     // im symbol2 die Amplitude durch den entsprechenden VLI Wert ersetzen
                lengthVLICoded += size+4;   // plus die 4 Bits fr die Lngeninfo
            }
            else // ein symbol1
                lengthVLICoded += 8;
        }
    }
    
    /**
     * Macht die Huffman Kodierung und schreibt die Lnge des Huffman Symbols
     * in <CODE>sizeNonZero</CODE>. Der Huffman kodierte Block wird in den
     * <CODE>HuffmanWriter</CODE> gesteckt, der der Methode bergeben wird.
     * Es muss auch je eine Tabelle fr die DC und eine fr die AC Komponenten
     * bergeben werden.
     * Bei AC Werten wird <CODE>sizeNonZero</CODE> & <CODE>zeroCount</CODE>
     * codiert, bei DC Werten nur <CODE>sizeNonZero</CODE>
     * @author Hari
     * @version 3 - 1.1.00
     * @param OurHuffmanTable dcHuffmanTable
     * @param OurHuffmanTable acHuffmanTable
     * @param HuffmanWriter hWriter
     */
    public void doHuffmanCoding(OurHuffmanTable dcHuffmanTable, OurHuffmanTable acHuffmanTable, HuffmanWriter hWriter) {
        int i, j;
        int val;
        int huffSize;
        RLESymbol symb;
        RLESymbol1 symb1;
        
        val = dcSymb1.sizeNonZero;  // bei DC nur die size (nicht die runlength) codieren
        huffSize = dcHuffmanTable.lengths[val];
        hWriter.writeSymbol(huffSize, dcHuffmanTable.symbols[val]);
        if (val > 0)
            hWriter.writeSymbol(val, dcSymb2.amplitude);
        dcSymb1.huffSize = (byte)huffSize;  // Gre nach Kodierung zurckschreiben, frs Histogramm spter

        symb1 = new RLESymbol1();  // das brauchen wir nur, weil der $*%!#? compiler denkt, er wre schlauer als ich.
        for (i=0; i<acSymbole.size(); i++) {
            symb = acSymbole.elementAt(i);
            if (symb instanceof RLESymbol1) {   // nur die symbol1 werden Huffman codiert
                symb1 = (RLESymbol1)symb;
                val = symb1.zeroCount*16 + symb1.sizeNonZero;  // runlength*16+size
                huffSize = acHuffmanTable.lengths[val];
                hWriter.writeSymbol(huffSize, acHuffmanTable.symbols[val]);
                symb1.huffSize = (byte)huffSize;  // Gre nach Kodierung zurckschreiben, frs Histogramm spter
            }
            else {  // ein symbol2
                int bits = symb1.sizeNonZero;  // das symbol1, das vor dem symbol2 kommt, steht noch in symb1
                hWriter.writeSymbol(bits, ((RLESymbol2)symb).amplitude);
            }
        }
    }

    /**
     * Gibt ein <CODE>Histogramm</CODE> zurck, in dem die Hufigkeiten der Bitlngen der
     * <CODE>RLESymbolfolge</CODE> nach der VLI Kodierung gespeichert sind
     * Es werden die Lngen der <CODE>RLESymbol2</CODE>'s gezhlt, plus das Feld 'sizeNonZero'
     * des dazugehrigen <CODE>RLESymbol1</CODE>
     * Die Methode ist sowohl auf DC- und AC Komponenten anwendbar. Ob es sich
     * um AC oder DC handelt, wird ber den Parameter <CODE>isDC</CODE>
     * angegeben (weil bei DC das (0,0) Symbol mitzhlt, bei AC bedeutet es EOB)
     * @author Hari
     * @version 2 - 2.1.00
     * @param boolean
     * @returns Histogramm
    */
    public Histogramm getVLIHistogramm() {
        Histogramm histo;
        int i;
        int bitSize;
        RLESymbol symb;
        
        histo = new Histogramm(16);  // 11 bits fr VLI und 4 fr Lnge, also 0..15
        
        // dc
        bitSize = 4+dcSymb1.sizeNonZero;  // die Lngeninfo zur VLI Lnge dazurechnen
        histo.werte[bitSize]++;
        
        // ac
        for (i=0; i<acSymbole.size(); i++) {
            symb = acSymbole.elementAt(i);
            if (symb instanceof RLESymbol1) {
                RLESymbol1 symb1 = (RLESymbol1)symb;
                bitSize = symb1.sizeNonZero;
                if (bitSize!=0 || (symb1.zeroCount!=0&&symb1.zeroCount!=15)) {  // EOB und (15,0) bei AC nicht mitzhlen
                    bitSize += 4;  // die Lngeninfo zur VLI Lnge dazurechnen
                    histo.werte[bitSize]++;
                }
            }
        }
        return histo;
    }

    /**
     * Gibt ein <CODE>Histogramm</CODE> zurck, in dem die Hufigkeiten der Bitlngen der
     * <CODE>RLESymbolfolge</CODE> nach der Huffman Kodierung gespeichert sind
     * Es werden die Lngen der Huffman Codes der <CODE>RLESymbol1</CODE>'s gezhlt
     * (Die Lngen stehen nach <CODE>doHuffmanCoding()</CODE> im Feld
     * <CODE>sizeNonZero</CODE> des Symbols)
     * Die Methode ist fr AC sowie fr DC Komponenten geeignet!
     * Aufgerufen wird sie, nachdem <CODE>doHuffmanCoding()</CODE> aufgerufen wurde
     * @author Hari
     * @version 1
     * @returns Histogramm
     */
    public Histogramm getHuffmanHistogramm() {
        Histogramm histo;
        int i;
        int bitSize;
        RLESymbol symb;
        
        histo = new Histogramm(17);  // Huffman Code ist auf 16 bits beschrnkt, also brauchen wir den bereich 0-16

        // dc
        bitSize = dcSymb1.huffSize;
        histo.werte[bitSize]++;
        
        // ac
        for (i=0; i<acSymbole.size(); i++) {
            symb = acSymbole.elementAt(i);
            if (symb instanceof RLESymbol1) {
                bitSize = ((RLESymbol1)symb).huffSize;
                histo.werte[bitSize]++;
            }
        }
        return histo;
    }
}    